from scipy.interpolate import interp1d
from functools import lru_cache
import logging

logger = logging.getLogger(__name__)

class ExtrapolateInterpolate:

    def __init__(self, x, y):
        """
        Interpolate and extrapolate a 1-D function.

        `x` and `y` are arrays of values used to approximate a function. This class returns a function whose call
        method uses interpolation and extrapolation to find the new points. The extrapolation is completed by
        choosing the last known data point (either maximum point in list, or minimum)
        :param x: array like X axis
        :param y: array like Y axis. The length of y must be the same as the length of x
        """
        self.x = x
        self.y = y

        if len(self.x) != len(self.y):
            raise ValueError("The x axis and y axis must be of the same size.")
        elif len(self.x)==0 or len(self.y)==0:
            raise ValueError("Length of x or y can not be 0.")

    def __call__(self, point):
        """
        Function which extrapolates and interpolates from known data. Use of linear interpolation between known
        points, and last known data point for extrapolation.
        :param point (int): Cost variable to be extrapolated/interpolated.
        :return (int): Returns extrapolated/interpolated cost of cost variable
        """
        if point <= min(self.x):
            return self.y.iloc[0]
        elif point >= max(self.x):
            return self.y.iloc[-1]
        else:
            interp = interp1d(self.x, self.y)
            return interp(point)

    def min_max_extrapolate(self, point):
        if point <= min(self.x):
            return self.y.iloc[0]
        elif point >= max(self.x):
            return self.y.iloc[-1]
        else:
            index_position = self.x[self.x == point].index[0]
            return self.y.iloc[index_position]



import sys

"""
File name: scenario_modifier
Date created: 03/03/2019
Feature: # Feature to enable the user to overwrite the scenario file.
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"

def overwrite_scenario_file(scenario_file):
    sys.modules['elecsim'].scenario.scenario_data=scenario_file



import logging

from elecsim.agents.generation_company.gen_co import GenCo
from elecsim.plants.plant_type.non_fuel_plant import NonFuelPlant

logger = logging.getLogger(__name__)

"""
File name: plant_capacity
Date created: 30/12/2018
Feature: # Functionality that aggregates all of the power plants in operation.
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"


class WorldPlantCapacity:
    def __init__(self, model):
        self.model = model

    def get_current_total_plant_capacity(self):
        plant_list = self.get_power_plants_running_in_current_year()
        total_capacity = self._calculate_total_capacity(plant_list)
        return total_capacity

    def get_renewable_plants(self):
        all_plants = self.get_power_plants_running_in_current_year()
        renewable_plants = [plant for plant in all_plants if isinstance(plant, NonFuelPlant)]
        return renewable_plants

    def get_renewable_by_type(self, type):
        renewable_plants = self.get_renewable_plants()
        renewable_type = [onshore_plant for onshore_plant in renewable_plants if onshore_plant.plant_type == type]
        return renewable_type

    def get_reference_year_total_capacity(self, reference_year):
        plant_list = self.get_power_plants_running_in_reference_year(reference_year)
        for i in plant_list:
            logger.debug(i.name)
        total_capacity = self._calculate_total_capacity(plant_list)
        return total_capacity

    def _calculate_total_capacity(self, plant_list):
        sumation = sum(plant.capacity_mw for plant in plant_list)
        return sumation

    def get_power_plants_running_in_current_year(self):
        plant_list = [plant for agent in self.model.schedule.agents if isinstance(agent, GenCo) for plant in
                      agent.plants if plant.is_operating is True]

        return plant_list

    def get_power_plants_running_in_reference_year(self, reference_year):
        plant_list = [plant for agent in self.model.schedule.agents if isinstance(agent, GenCo) for plant in
                      agent.plants if
                      plant.construction_year + plant.pre_dev_period + plant.construction_period < reference_year < plant.construction_year + plant.operating_period + plant.pre_dev_period + plant.construction_period]
        return plant_list



import functools
import logging
# from linetimer import CodeTimer
import numpy as np
import pandas as pd

# from elecsim.scen_error.scenario_data import wind_capacity_factor, solar_capacity_factor, historical_hourly_demand, \
#     segment_demand, load_duration_curve_diff, hydro_capacity_factor
import elecsim.scenario.scenario_data

logger = logging.getLogger(__name__)
"""
File name: solar_capacity_factor
Date created: 27/12/2018
Feature: # Calculates the average capacity factor per demand segment
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"


@functools.lru_cache(maxsize=1024)
def get_capacity_factor(market_time_splices, renewable_type, demand_hour):
    renewable_type = renewable_type.lower()
    if market_time_splices == 1:
        historical_demand = elecsim.scenario.scenario_data.historical_hourly_demand
        capacity_data = get_capacity_data(renewable_type)
    elif market_time_splices > 1:
        historical_demand = elecsim.scenario.scenario_data.historical_hourly_demand
        capacity_data = get_multiple_capacity_data(renewable_type)
    else:
        raise ValueError("market_time_slices must be bigger than 1")
    if renewable_type in ['onshore', 'offshore', 'pv']:
        if market_time_splices == 1:
            capacity_factor = segment_capacity_data_by_load_curve(renewable_type, market_time_splices)
            logger.info(segment_capacity_data_by_load_curve.cache_info())
            capacity_factor_value = get_capacity_factor_value_for_segment(capacity_factor, demand_hour, renewable_type)
        else:
            # capacity_factor = segment_multiple_capacity_data_by_load_curve(capacity_data, historical_demand, model)
            capacity_factor_value = get_multi_capacity_factor_value_for_segment(renewable_type, demand_hour)
    else:
        capacity_factor_value = capacity_data
    return capacity_factor_value


def get_multiple_capacity_data(renewable_type):
    capacity_factor_data = elecsim.scenario.scenario_data.multi_year_data_scaled
    if renewable_type in ["onshore", 'offshore']:
        capacity_data = capacity_factor_data[capacity_factor_data.data_type == renewable_type]
    elif renewable_type == "pv":
        capacity_data = capacity_factor_data[capacity_factor_data.data_type == 'solar']
    elif renewable_type == 'hydro':
        capacity_data = elecsim.scenario.scenario_data.hydro_capacity_factor
    elif renewable_type == "nuclear":
        capacity_data = elecsim.scenario.scenario_data.nuclear_capacity_factor
    else:
        raise ValueError("Calculating demand factor can only be done for Onshore, Offshore or PV power generators.")
    return capacity_data


def get_capacity_data(renewable_type):
    if renewable_type in ["onshore", 'offshore']:
        capacity_data = elecsim.scenario.scenario_data.wind_capacity_factor[['time', renewable_type]]
    elif renewable_type == "pv":
        capacity_data = elecsim.scenario.scenario_data.solar_capacity_factor
    elif renewable_type == 'hydro':
        capacity_data = elecsim.scenario.scenario_data.hydro_capacity_factor
    elif renewable_type == "nuclear":
        capacity_data = elecsim.scenario.scenario_data.nuclear_capacity_factor
    else:
        raise ValueError("Calculating demand factor can only be done for Onshore, Offshore or PV power generators.")
    return capacity_data


@functools.lru_cache(maxsize=512)
def segment_capacity_data_by_load_curve(renewable_type, market_time_splices):
    historical_demand = elecsim.scenario.scenario_data.historical_hourly_demand
    capacity_data = get_capacity_data(renewable_type)

    demand_capacity = capacity_data.join(historical_demand, how='inner').dropna()
    demand_capacity = demand_capacity[elecsim.scenario.scenario_data.segment_demand[-1] < demand_capacity.demand]
    demand_capacity = demand_capacity[elecsim.scenario.scenario_data.segment_demand[0] > demand_capacity.demand]
    cut_demand = pd.cut(demand_capacity.demand, 20* market_time_splices)
    capacity_factor_grouped = demand_capacity.groupby(cut_demand)[renewable_type]
    capacity_factor = capacity_factor_grouped.apply(np.mean)
    capacity_factor = capacity_factor.to_frame()
    capacity_factor['diff'] = elecsim.scenario.scenario_data.load_duration_curve_diff['hours'].sort_values(ascending=False).values
    return capacity_factor


def get_capacity_factor_value_for_segment(capacity_factor, demand_hour, renewable_type):
    subsetted_dataframe = capacity_factor[capacity_factor['diff'] >= demand_hour].tail(1)
    capacity_factor_value = float(subsetted_dataframe[renewable_type].values)
    return capacity_factor_value


def get_multi_capacity_factor_value_for_segment(renewable_type, demand_hour):
    capacity_factor = get_multiple_capacity_data(renewable_type)

    capacity_factor_value = capacity_factor[capacity_factor['demand_hour'] >= demand_hour].capacity_factor.head(1)
    capacity_factor_value = capacity_factor_value.values[0]

    # subsetted_dataframe = capacity_factor[capacity_factor['diff'] >= demand_hour].tail(1)
    # capacity_factor_value = float(subsetted_dataframe['capacity_factor'].values)
    # capacity_factor_value = subsetted_dataframe['capacity_factor'].values.astype(float)
    return capacity_factor_value


def get_capacity_factor_for_year(renewable_type):
    if renewable_type == "pv":
        renewable_type = "solar"
    capacity_factor_dat = elecsim.scenario.scenario_data.multi_year_data
    capacity_factor = multi_year_data_to_ldc(capacity_factor_dat, renewable_type)

    year_segment_capacity_factor = (pd.DataFrame(pd.cut(capacity_factor.capacity_factor, 20).apply(lambda x: x.right)).groupby("capacity_factor").first()).index.tolist()
    year_segment_hours = (pd.DataFrame(pd.cut(capacity_factor.hour, 20).apply(lambda x: x.right)).groupby("hour").first()).index.tolist()

    return year_segment_hours, year_segment_capacity_factor


def multi_year_data_to_ldc(capacity_factor_dat, renewable_type):
    renewable_capacity_factor = capacity_factor_dat[capacity_factor_dat.data_type == renewable_type]
    capacity_factor = renewable_capacity_factor.sort_values("capacity_factor", ascending=False)
    capacity_factor['hour'] = capacity_factor.counts.cumsum()
    return capacity_factor


functools.lru_cache(maxsize=512)
def segment_multiple_capacity_data_by_load_curve(capacity_data, historical_demand, model):

    demand_capacity = capacity_data.join(historical_demand, how='inner').dropna()
    demand_capacity = demand_capacity[model.demand.segment_consumption[-1] < demand_capacity.demand]
    demand_capacity = demand_capacity[model.demand.segment_consumption[0] > demand_capacity.demand]
    cut_demand = pd.cut(demand_capacity.demand, 20 * model.market_time_splices)
    capacity_factor_grouped = demand_capacity.groupby(cut_demand)['capacity_factor']
    capacity_factor = capacity_factor_grouped.apply(np.mean)
    capacity_factor = capacity_factor.to_frame().fillna(0)

    capacity_factor['diff'] = np.linspace(0.083, 8752.5, 20 * model.market_time_splices)

    # capacity_factor['diff'] = capacity_factor.reset_index()['demand'].apply(lambda x: x.left).astype(float).values
    # capacity_factor = capacity_factor.reset_index()
    # capacity_factor['diff'] = elecsim.scenario.scenario_data.load_duration_curve_diff['hours'].sort_values(ascending=False).values

    # capacity_factor_dat = elecsim.scenario.scenario_data.multi_year_data
    # load_factor = multi_year_data_to_ldc(capacity_factor_dat, 'load')
    #
    # capacity_factor['diff'] = pd.cut(load_factor['hour'], 20 * model.market_time_splices).sort_values(ascending=False).values
    return capacity_factor




import pandas as pd

import elecsim.scenario.scenario_data
from elecsim.data_manipulation.data_modifications.extrapolation_interpolate import ExtrapolateInterpolate
from elecsim.plants.plant_costs.estimate_costs.estimate_modern_power_plant_costs.predict_modern_plant_costs import PredictModernPlantParameters
from elecsim.plants.plant_registry import PlantRegistry
# from elecsim.scen_error.scenario_data import modern_plant_costs


class OldPlantCosts:
    """
    Class which takes LCOE values and plant_type of power plants from retrospective database and predicts
    more detailed cost parameters using the same proportions as the BEIS Power Plant Cost Database.
    Specifically uses 2018 power plants from BEIS Power Plant Cost Database.
    """
    hist_costs = elecsim.scenario.scenario_data.power_plant_historical_costs_long

    def __init__(self, year, plant_type, capacity):

        # Import historical LCOE data for power plants, and use to predict LCOE for current year based on linear
        # interpolation
        self.year = year
        self.plant_type = plant_type
        self.capacity = capacity
        self.hist_costs = self.hist_costs[self.hist_costs['Technology'].map(lambda x: x in plant_type)].dropna()

        if not all(self.hist_costs.capacity_range.str.contains(">0")):
            self.hist_costs = self.hist_costs[[pd.eval(f"{self.capacity}{j}") for j in self.hist_costs['capacity_range']]]

        self.estimated_historical_lcoe = ExtrapolateInterpolate(self.hist_costs.Year, self.hist_costs.lcoe)(year)
        self.discount_rate = self.hist_costs.Discount_rate.iloc[0]

        self.modern_costs = elecsim.scenario.scenario_data.modern_plant_costs[elecsim.scenario.scenario_data.modern_plant_costs['Type'].map(lambda x: x in plant_type)]
        minimum_year = self.find_smallest_year_available()

        self.estimated_modern_plant_parameters = PredictModernPlantParameters(self.plant_type, self.capacity, minimum_year).parameter_estimation()

        plant_object = PlantRegistry(self.plant_type).plant_type_to_plant_object()

        self.plant = plant_object(name="Modern Plant", plant_type=self.plant_type,
                                                         capacity_mw=self.capacity, construction_year=self.year,
                                                         **self.estimated_modern_plant_parameters)
        self.modern_lcoe = self.plant.calculate_lcoe(self.discount_rate)
        self.lcoe_scaler = self.estimated_historical_lcoe / self.modern_lcoe

    def find_smallest_year_available(self):
        """
        Method which takes the modern cost BEIS database of power plants, and finds the earliest year
        that data for specified power plant type exists. For example, only returns data on Coal power plants from 2025
        as only this data is provided in the BEIS datafile
        :return: Int containing smallest year available.
        """
        # available_years = self.modern_costs[['Constr_cost-Medium _2018','Constr_cost-Medium _2020', 'Constr_cost-Medium _2025']]
        available_years = self.modern_costs.filter(regex="Constr_cost-Medium _")
        columns_with_no_nan = available_years[available_years.columns[~available_years.isnull().all()]].columns
        years_with_no_nan = [s for s in columns_with_no_nan]
        years_with_no_nan = [int(s.split("_")[2]) for s in years_with_no_nan]
        minimum_year_with_data = min(years_with_no_nan)

        return minimum_year_with_data

    def get_params_to_scale(self):
        self.params_to_ignore = ['pre_dev_period', 'operating_period', 'construction_period', 'efficiency',
                            'average_load_factor', 'construction_spend_years', 'pre_dev_spend_years']
        self.dict_to_ignore = {key: self.estimated_modern_plant_parameters[key] for key in
                          self.estimated_modern_plant_parameters
                          if key in self.params_to_ignore}






def closest_row(dataframe, column, value):
    """
    Function which takes a dataframe and returns the row that is closest to the specified value of the specified column.
    :param dataframe: Dataframe object
    :param column: String which matches to a column in the dataframe in which you would like to find the closest value of.
    :param value: Value to find the closest row to.
    :return: Returns row that is closest to the value of the selected column of the dataframe
    """
    sort = dataframe.iloc[(dataframe[column]-value).abs().argsort()[:1]]
    return sort




from functools import lru_cache

from elecsim.data_manipulation.data_modifications.extrapolation_interpolate import ExtrapolateInterpolate
import elecsim.scenario.scenario_data
import elecsim.scenario.scenario_data
import elecsim.scenario.scenario_data
"""
File name: availability_factor_calculations
Date created: 13/01/2019
Feature: #Enter feature description here
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"


@lru_cache(1024)
def get_availability_factor(plant_type, construction_year):
    simplified_type = get_plant_type_for_data(plant_type)
    if plant_type in ['CCGT', "OCGT", "Coal", 'CHP']:
        availability_factor_series = elecsim.scenario.scenario_data.historical_availability_factor[(elecsim.scenario.scenario_data.historical_availability_factor.plant_type==simplified_type)][['Year','capacity_factor']]
        availability_factor = ExtrapolateInterpolate(availability_factor_series.Year, availability_factor_series.capacity_factor)(construction_year)
    else:
        availability_factor = elecsim.scenario.scenario_data.fuel_plant_availability
    return availability_factor


def get_plant_type_for_data(plant_type):
    if plant_type == "CCGT":
        plant = 'combined_cycle'
    elif plant_type == "OCGT":
        plant = 'combined_cycle'
    elif plant_type == "Coal":
        plant = 'Coal'
    elif plant_type == "CHP":
        plant = "CHP"
    else:
        return elecsim.scenario.scenario_data.fuel_plant_availability
    return plant




import logging
from itertools import zip_longest
from math import isnan
from functools import lru_cache

import pandas as pd
import numpy as np
import elecsim.scenario.scenario_data
from elecsim.constants import ROOT_DIR
from elecsim.data_manipulation.data_modifications.extrapolation_interpolate import ExtrapolateInterpolate
from elecsim.plants.fuel.fuel_registry.fuel_registry import fuel_registry, plant_type_to_fuel
from elecsim.role.plants.costs.plant_cost_calculation import PlantCostCalculations

logger = logging.getLogger(__name__)

"""
File name: fuel_lcoe_calculation
Date created: 18/12/2018
Feature: #Enter feature description here
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"


class FuelPlantCostCalculations(PlantCostCalculations):
    def __init__(self, plant_type, capacity_mw, construction_year, average_load_factor, efficiency, pre_dev_period,
                 construction_period, operating_period, pre_dev_spend_years, construction_spend_years,
                 pre_dev_cost_per_mw, construction_cost_per_mw, infrastructure, fixed_o_and_m_per_mw,
                 variable_o_and_m_per_mwh, insurance_cost_per_mw, connection_cost_per_mw):
        super().__init__(capacity_mw=capacity_mw, construction_year=construction_year,
                         average_load_factor=average_load_factor, pre_dev_period=pre_dev_period,
                         construction_period=construction_period, operating_period=operating_period,
                         pre_dev_spend_years=pre_dev_spend_years, construction_spend_years=construction_spend_years,
                         pre_dev_cost_per_mw=pre_dev_cost_per_mw, construction_cost_per_mw=construction_cost_per_mw,
                         infrastructure=infrastructure, fixed_o_and_m_per_mw=fixed_o_and_m_per_mw,
                         variable_o_and_m_per_mwh=variable_o_and_m_per_mwh, insurance_cost_per_mw=insurance_cost_per_mw,
                         connection_cost_per_mw=connection_cost_per_mw)
        self.plant_type = plant_type
        self.efficiency = efficiency
        # Finds fuel plant_type of power plant eg. CCGT power plant plant_type returns gas.
        fuel_string = plant_type_to_fuel(plant_type, self.construction_year)
        # Fuel object, containing information on fuel.
        self.fuel = fuel_registry(fuel_string)
        self.fuel_string = fuel_string

    def calculate_lcoe(self, discount_rate):
        """
        Function which calculates the levelised cost of electricity for this power plant instance at a
        specified discount rate.

        :param discount_rate: The discount rate that is used for the calculation of the levelised cost of electricity.
        :return: Returns LCOE value for power plant
        """

        # Calculations of capital expenditure, operating expenditure, total expected electricity expenditure and plant_type cost
        # This is used to estimate a LCOE price.
        elec_gen, total_costs = self.calculate_total_costs()
        # Costs discounted using discount_rate variable.
        disc_costs = self._discount_data(total_costs, discount_rate)
        disc_elec = self._discount_data(elec_gen, discount_rate)

        # Sum total costs over life time of plant
        disc_total_costs = sum(disc_costs)
        disc_total_elec = sum(disc_elec)

        # LCOE calculated
        lcoe = disc_total_costs / disc_total_elec

        return lcoe

    def calculate_total_costs(self, estimated_fuel_prices=None, estimated_carbon_prices=None):
        capex = self._capex()
        opex = self._opex_cost()
        elec_gen = self._electricity_generated()
        fuel_costs = self._fuel_costs(elec_gen)
        carbon_costs = self._carbon_costs()
        total_costs = self._total_costs(capex, opex, fuel_costs, carbon_costs)

        return elec_gen, total_costs

    def _total_costs(self, capex, opex, fuel_costs, carbon_costs):
        """
        Function which uses addition to calculate total costs from capital expenditure, operating expenditure, fuel costs,
        and carbon costs for plant_type costs over the lifetime of the power plant.

        :param capex: Capital expenditure per year
        :param opex: Operating expenditure per year
        :param fuel_costs: _fuel_costs per year
        :return: Total costs over lifetime of power plant
        """

        # Addition of operating expenditure and plant_type costs, followed by operating expenditure, plant_type costs and capital expenditure.
        opex_fuel = [sum(x) for x in zip_longest(opex, fuel_costs, fillvalue=0)]

        opex_fuel_remove_0 = opex_fuel[int(self.construction_period + self.pre_dev_period):]
        capex.extend(opex_fuel_remove_0)
        sum_of_opex_fuel_capex = capex.copy()
        sum_of_carbon_opex_fuel_capex = [sum(x) for x in zip(sum_of_opex_fuel_capex, carbon_costs)]

        return sum_of_carbon_opex_fuel_capex

    def _fuel_costs(self, electricity_generated, fuel_price=None):
        """
        Calculates the plant_type costs per year based on plant efficiency, electricity generated and endogenous gas prices
        :param electricity_generated: Electricity generated per year
        :return: Returns estimated cost of plant_type per year
        """
        beginning_year_operation = self.construction_year
        end_of_lifetime_year = int(beginning_year_operation) + int(self.operating_period) + int(
            self.pre_dev_period + self.construction_period)
        years_of_plant_operation = range(int(beginning_year_operation), end_of_lifetime_year)

        if fuel_price is None:
            this_fuel_price = self.fuel.fuel_price[self.fuel.fuel_price.Fuel == self.fuel.fuel_type].dropna()
        else:
            this_fuel_price = fuel_price

        fuel_extrapolation = ExtrapolateInterpolate(this_fuel_price.Year, this_fuel_price.value)
        fuel_price = [(float(fuel_extrapolation(i)) * elec_gen) / self.efficiency for i, elec_gen in
                      zip(years_of_plant_operation, electricity_generated)]

        return fuel_price

    def _carbon_emitted(self):
        """
        Calculates projected tonnes of CO2 emitted by power plant per year
        :return: A list containing tonnes of CO2 emitted per year
        """
        carbon_emitted = [self.fuel.mwh_to_co2e_conversion_factor * (elec_gen / self.efficiency) for elec_gen in
                          self._electricity_generated()]
        return carbon_emitted

    def _carbon_costs(self, carbon_price=None):
        """
        Costs of carbon emissions based on carbon tax
        :return: Projected carbon costs per year in a list.
        """

        if carbon_price is not None:
            carbon_taxation_years = carbon_price
        else:
            carbon_cost = self.get_BEIS_carbon_price()

            year_of_beginning_operation = self.construction_year + self.pre_dev_period + self.construction_period
            carbon_taxation_years = carbon_cost[carbon_cost.year.between(int(self.construction_year), int(
                year_of_beginning_operation + self.operating_period - 1))]

        carbon_emitted = self._carbon_emitted()

        carbon_costs = [carbon_tax * carb_emit for carbon_tax, carb_emit in
                        zip(list(carbon_taxation_years.price), carbon_emitted)]

        return carbon_costs



    def _carbon_cost_total(self, carbon_costs):
        """
        Calculates the total cost of carbon over the lifetime of the power plant.
        :return: total carbon costs
        """
        carbon_costs_total = sum(carbon_costs)

        return carbon_costs_total

    def calculate_short_run_marginal_cost(self, model, genco, fuel_price=None, co2_price=None):
        """
        Calculates the short run marginal cost for a fuel power plant
        :param model: Model containing information such as current year
        :param genco: Generation company object that requires short run marginal cost. Used to use genco price of fuel.
        :return: returns marginal cost to burn 1MWh of fuel.
        """

        modifier = 0
        if self.plant_type == 'CCGT':
            modifier = genco.gas_price_modifier
        elif self.plant_type == 'Coal':
            modifier = genco.coal_price_modifier

        fuel_cost = self.get_fuel_price(fuel_price, model.year_number, modifier)
        co2_cost = self.get_co2_price(co2_price, model)
        marginal_cost = self.variable_o_and_m_per_mwh + fuel_cost + co2_cost
        if isnan(marginal_cost):
            logger.debug("Marginal cost is nan. Variable cost: {}, Fuel cost: {}, CO2 Cost: {}, plant type: {}".format(
                self.variable_o_and_m_per_mwh, fuel_cost, co2_cost, self.plant_type))
            raise ValueError("Marginal cost is nan. Variable cost: {}, Fuel cost: {}, CO2 Cost: {}, plant type: {}".format(
                self.variable_o_and_m_per_mwh, fuel_cost, co2_cost, self.plant_type))

        return marginal_cost

    def get_co2_price(self, co2_price, model):
        if co2_price is None:
            # co2_cost = self.fuel.mwh_to_co2e_conversion_factor * (1 / self.efficiency) * \
            #            carbon_cost[carbon_cost.year == model.year_number - 1].price.iloc[0]

            co2_cost = self.fuel.mwh_to_co2e_conversion_factor * (1 / self.efficiency) * get_carbon_cost_in_year(
                model.year_number)
        else:
            co2_cost = self.fuel.mwh_to_co2e_conversion_factor * (1 / self.efficiency) * co2_price
        return co2_cost

    def get_fuel_price(self, fuel_price, year_number, modifier):
        if fuel_price is None:
            # logger.debug("fuel_price_is_none calculating from data")
            # fuel_cost = (self.fuel.fuel_price[self.fuel.fuel_price.Year == model.year_number - 1].value.iloc[
            #                  0] + modifier) / self.efficiency


            # fuel_price = self.fuel.fuel_price[self.fuel.fuel_price.Year == model.year_number - 1].value.iloc[0] + modifier
            fuel_price = _query_fuel_price_for_year(self.fuel_string, year_number) + modifier
            logger.debug("_query_fuel_price_for_year: {}".format(_query_fuel_price_for_year.cache_info()))

            fuel_cost = fuel_price / self.efficiency
        else:
            # logger.debug("fuel_price_is_given calculating from parenthesis")
            fuel_cost = fuel_price / self.efficiency
        return fuel_cost

    def get_BEIS_carbon_price(self):
        carbon_price_scenario = [18.00, 19.42, 20.83, 22.25, 23.67, 25.08, 26.50, 27.92, 29.33, 30.75, 32.17, 33.58,
                                 35.00, 43.25, 51.50, 59.75, 68.00, 76.25, 84.50, 92.75, 101.00, 109.25, 117.50,
                                 125.75, 134.00, 142.25, 150.50, 158.75, 167.00, 175.25, 183.50, 191.75,
                                 200.00]  # Forecast used from BEIS Electricity Generation Report - Page 10 - Includes forecast for carbon tax and EU ETS
        carbon_data = {'year': [str(i) for i in range(2019, (2019 + len(carbon_price_scenario)))],
                       'price': carbon_price_scenario}
        carbon_price_scenario_df = pd.DataFrame(carbon_data)
        historical_carbon_price = pd.read_csv(
            ROOT_DIR + '/data/processed/carbon_price/uk_carbon_tax_historical.csv')
        carbon_cost = historical_carbon_price.append(carbon_price_scenario_df, sort=True)
        carbon_cost.year = pd.to_numeric(carbon_cost.year)
        return carbon_cost


# lru_cache(maxsize=128)
def calculate_year_carbon_price():
    # carbon_price_scenario = elecsim.scenario.scenario_data.carbon_price_scenario
    # # logger.debug("Calculating short_run_marginal_cost")
    # # Join historical and future carbon prices into dataframe for simulation purposes
    # carbon_data = {'year': [str(i) for i in range(2019, (2019 + len(carbon_price_scenario)))],
    #                 'price': carbon_price_scenario}
    # carbon_price_scenario_df = pd.DataFrame(carbon_data)
    # historical_carbon_price = pd.read_csv(ROOT_DIR + '/data/processed/carbon_price/uk_carbon_tax_historical.csv')
    # carbon_cost = historical_carbon_price.append(carbon_price_scenario_df, sort=True)
    # carbon_cost.year = pd.to_numeric(carbon_cost.year)
    carbon_cost = elecsim.scenario.scenario_data.carbon_price_all_years
    return carbon_cost


@lru_cache(maxsize=1000)
def get_carbon_cost_in_year(year_number):
    # carbon_cost = carbon_cost[carbon_cost.year == year_number - 1].price.iloc[0]
    carbon_cost = elecsim.scenario.scenario_data.carbon_price_all_years

    carbon_cost = carbon_cost.loc[np.in1d(carbon_cost['year'], [year_number]), 'price'].iloc[0]
    return carbon_cost

@lru_cache(1024)
def _query_fuel_price_for_year(fuel_string, year_number):
    fuel = fuel_registry(fuel_string)
    fuel_price = fuel.fuel_price.loc[np.in1d(fuel.fuel_price['Year'], [year_number]), 'value'].iloc[0]
    return fuel_price



import logging
from functools import lru_cache
from itertools import chain
# from linetimer import CodeTimer
import pandas as pd

from elecsim.market.electricity.bid import Bid
import elecsim.scenario.scenario_data

logger = logging.getLogger(__name__)



"""power_exchange.py: Functionality to run power exchange"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "Alexander@Kell.es"


class PowerExchange:

    def __init__(self, model):
        """
        Power exchange agent which contains functionality to tender and respond to bids.
        :param model: Model in which the agents are contained in.
        """
        self.model = model
        self.hold_duration_curve_prices = []
        self.price_duration_curve = pd.DataFrame(columns=["year", "segment_hour", "segment_demand", "accepted_price"])

        self.stored_bids = {}
        self.stored_ordered_bids = {}

    def tender_bids(self, segment_hours, segment_demand, predict=False):
        """
        Function which iterates through the generator companies, requests their bids, orders them in order of price,
        and accepts bids.
        :param agents: All agents from simulation model.
        :param segment_hours: Value for number of hours particular electricity generation is required.
        :param segment_demand: Size of electricity consumption required.
        :param predict: Boolean that states whether the bids being tendered are for predicting price duration curve or whether it is for real bids.
        :return: None
        """

        self.hold_duration_curve_prices = []

        agent = self.model.schedule.agents
        generator_companies = [x for x in agent if hasattr(x, 'plants')]  # Selection of generation company agents

        for gen_co in generator_companies:
            for plant in gen_co.plants:
                plant.capacity_fulfilled = dict.fromkeys(segment_hours, 0)

        for segment_hour, segment_demand in zip(segment_hours, segment_demand):
            bids = []

            for generation_company in generator_companies:
                bids.append(generation_company.calculate_bids(segment_hour, predict))
                # logger.info(generation_company.calculate_bids.cache_info())

            sorted_bids = self._sort_bids(bids)
            if predict is False:

                logger.debug("bids len: {}".format(len(sorted_bids)))
                # logger.info("total capacity of bids: {}".format(sum(bid.capacity_bid for bid in sorted_bids)))

            accepted_bids = self._respond_to_bids(sorted_bids, segment_hour, segment_demand)

            logger.debug("segment hour: {}".format(segment_hour))
            highest_bid = self._accept_bids(accepted_bids)
            # highest_bid = max(bid.price_per_mwh for bid in accepted_bids)

            self._create_load_duration_price_curve(segment_hour, segment_demand, highest_bid)

        self.price_duration_curve = pd.DataFrame(self.hold_duration_curve_prices)
        if predict:
            self.price_duration_curve = self.price_duration_curve[(self.price_duration_curve.year == self.model.year_number) & (self.price_duration_curve.day == self.model.step_number)]
            logger.debug("predicted self.price_duration_curve: \n{}".format(self.price_duration_curve))
        else:
            self.price_duration_curve = self.price_duration_curve[(self.price_duration_curve.year == self.model.year_number) & (self.price_duration_curve.day == self.model.step_number)]
            logger.debug("actual self.price_duration_curve: \n{}".format(self.price_duration_curve))
        # return self.price_duration_curve[self.price_duration_curve.year == self.model.year_number].accepted_price.mean()
        return self.price_duration_curve[self.price_duration_curve.year == self.model.year_number].accepted_price#.mean()



    # def tender_bids(self, segment_hours, segment_demand, predict=False):
    #     """
    #     Function which iterates through the generator companies, requests their bids, orders them in order of price,
    #     and accepts bids.
    #     :param agents: All agents from simulation model.
    #     :param segment_hours: Value for number of hours particular electricity generation is required.
    #     :param segment_demand: Size of electricity consumption required.
    #     :param predict: Boolean that states whether the bids being tendered are for predicting price duration curve or whether it is for real bids.
    #     :return: None
    #     """
    #     agent = self.model.schedule.agents
    #     generator_companies = [x for x in agent if hasattr(x, 'plants')]  # Selection of generation company agents
    #
    #     if self.model.beginning_of_year:
    #         self.hold_duration_curve_prices = []
    #
    #
    #     save = dict.fromkeys(segment_hours, 0)
    #     for gen_co in generator_companies:
    #         for plant in gen_co.plants:
    #             # plant.capacity_fulfilled = dict.fromkeys(segment_hours, 0)
    #             plant.capacity_fulfilled = save.copy()
    #
    #     for segment_hour, segment_demand in zip(segment_hours, segment_demand):
    #         bids = []
    #
    #         if segment_hour not in self.stored_bids or predict is False:
    #             sorted_bids = self._calculate_all_bids(bids, generator_companies, predict, segment_hour)
    #             self.stored_bids[segment_hour] = sorted_bids
    #         else:
    #             sorted_bids = self.stored_bids[segment_hour]
    #             if self.model.last_added_plant:
    #                 # logger.info(self.model.last_added_plant_bids)
    #
    #                 sorted_bids.append(self.model.last_added_plant[segment_hour])
    #
    #         accepted_bids = self._respond_to_bids(sorted_bids, segment_hour, segment_demand)
    #
    #         self.stored_ordered_bids[segment_hour] = accepted_bids
    #
    #         logger.debug("segment hour: {}".format(segment_hour))
    #
    #         self._accept_bids(accepted_bids)
    #         highest_bid = max(bid.price_per_mwh for bid in self.stored_ordered_bids[segment_hour] if bid.bid_accepted is True)
    #         self._create_load_duration_price_curve(segment_hour, segment_demand, highest_bid)
    #
    #     self.price_duration_curve = pd.DataFrame(self.hold_duration_curve_prices)
    #     if predict:
    #         self.price_duration_curve = self.price_duration_curve[(self.price_duration_curve.year == self.model.year_number)]
    #         logger.debug("predicted self.price_duration_curve: {}".format(self.price_duration_curve))
    #     else:
    #         self.price_duration_curve = self.price_duration_curve[(self.price_duration_curve.year == self.model.year_number) & (self.price_duration_curve.day == self.model.step_number)]
    #         logger.info("actual self.price_duration_curve: {}".format(self.price_duration_curve))
    #
    #     return self.price_duration_curve[self.price_duration_curve.year == self.model.year_number].accepted_price.mean()

    def _calculate_all_bids(self, bids, generator_companies, predict, segment_hour):
        # for generation_company in generator_companies:
        #     bids.append(generation_company.calculate_bids(segment_hour, predict))

        bids = [generation_company.calculate_bids(segment_hour, predict) for generation_company in generator_companies]
            # logger.info(generation_company.calculate_bids.cache_info())
        sorted_bids = self._sort_bids(bids)
        return sorted_bids

    def _create_load_duration_price_curve(self, segment_hour, segment_demand, accepted_price):
        segment_price_data = {
                'year': self.model.year_number,
                'day': self.model.step_number,
                'segment_hour': segment_hour,
                'segment_demand': segment_demand,
                'accepted_price': accepted_price
            }

        self.hold_duration_curve_prices.append(segment_price_data)

    @staticmethod
    def _accept_bids(accepted_bids):
        highest_accepted_bid = accepted_bids[-1].price_per_mwh

        for bid in accepted_bids:
            # logger.debug("bid price: {}, plant name: {}, plant capacity: {}".format(bids.price_per_mwh, bids.plant.name, bids.plant.capacity_mw))
            bid.price_per_mwh = highest_accepted_bid

        return highest_accepted_bid

    @staticmethod
    def _sort_bids(bids, attribute="price_per_mwh"):
        """
        Sorts bids in order of price
        :param bids: Bid objects
        :return: Return bids in order of price
        """
        bids = list(chain.from_iterable(bids))
        sorted_bids = sorted(bids, key=lambda x: getattr(x, attribute))
        return sorted_bids


    def _respond_to_bids(self, bids, segment_hour, capacity_required):
        """
        Response to bids based upon price and capacity required. Accepts bids in order of cheapest generator.
        Continues to accept bids until capacity is met for those hours.
        :param bids: Bid objects.
        :param capacity_required: Capacity required for this segment.
        :return: Returns a list of bids which have been accepted.
        """
        accepted_bids = []

        for bid in bids:
            if capacity_required > 0 and capacity_required > bid.capacity_bid:
                bid.accept_bid(segment_hour)
                capacity_required -= bid.capacity_bid
                accepted_bids.append(bid)
                logger.debug('bid ACCEPTED: price: {}, year: {}, capacity required: {}, capacity: {}, capacity_bid: {}, type: {}, name {}'.format(bid.price_per_mwh, bid.plant.construction_year, capacity_required, bid.plant.capacity_mw, bid.capacity_bid, bid.plant.plant_type,  bid.plant.name))
            elif bid.capacity_bid > capacity_required > 0:
                bid.partially_accept_bid(segment_hour, capacity_required)
                capacity_required = 0
                accepted_bids.append(bid)
                logger.debug('bid PARTIALLY ACCEPTED: price: {}, year: {}, capacity required: {}, capacity: {}, capacity_bid: {}, type: {}, name {}'.format(bid.price_per_mwh, bid.plant.construction_year, capacity_required, bid.plant.capacity_mw, bid.capacity_bid, bid.plant.plant_type,  bid.plant.name))
            else:
                bid.reject_bid(segment_hour=segment_hour)
                logger.debug('bid REJECTED: price: {}, year: {}, capacity required: {}, capacity: {}, capacity_bid: {}, type: {}, name {}'.format(bid.price_per_mwh, bid.plant.construction_year, capacity_required, bid.plant.capacity_mw, bid.capacity_bid, bid.plant.plant_type,  bid.plant.name))
        if capacity_required > 0:
            accepted_bids.append(Bid(None, None, segment_hour, 0, elecsim.scenario.scenario_data.lost_load, self.model.year_number))
        return accepted_bids



#     def _respond_to_bids(self, bids, segment_hour, capacity_required):
#         """
#         Response to bids based upon price and capacity required. Accepts bids in order of cheapest generator.
#         Continues to accept bids until capacity is met for those hours.
#         :param bids: Bid objects.
#         :param capacity_required: Capacity required for this segment.
#         :return: Returns a list of bids which have been accepted.
#         """
#
#         if segment_hour not in self.stored_ordered_bids:
#             accepted_bids = NoHistoryBidResponder(bids, capacity_required, segment_hour).get_accepted_bids()
#         elif self.model.last_added_plant_bids is None:
#             accepted_bids = self.stored_ordered_bids[segment_hour]
#         elif segment_hour in self.stored_ordered_bids:
#             accepted_bids = HistoryBidResponder(self.stored_ordered_bids[segment_hour], capacity_required, segment_hour).get_accepted_bids(new_bid=self.model.last_added_plant_bids[segment_hour])
#         else:
#             accepted_bids = NoHistoryBidResponder(bids, capacity_required, segment_hour).get_accepted_bids()
#
#         return accepted_bids
#
#
# class BidResponder:
#
#     def __init__(self, bids, capacity_required, segment_hour):
#         self.capacity_required = capacity_required
#         self.segment_hour = segment_hour
#         self.bids = bids
#
#
# class HistoryBidResponder(BidResponder):
#
#     def __init__(self, bids, capacity_required, segment_hour):
#         super().__init__(bids=bids, capacity_required=capacity_required, segment_hour=segment_hour)
#
#     def get_accepted_bids(self, new_bid):
#         # logger.info("accepted_price: {}".format(self.bids[0].price_per_mwh))
#         # logger.info("new_bid: {}".format(new_bid))
#
#         if new_bid.price_bid < self.bids[0].price_per_mwh:
#             capacity_undercut = new_bid.capacity_bid
#             while capacity_undercut > 0:
#                 self.bids = PowerExchange._sort_bids([self.bids], "price_bid")
#                 most_expensive_bid_capacity = self.bids[-1].capacity_bid
#
#                 if self.bids[0].price_per_mwh == elecsim.scenario.scenario_data.lost_load:
#                     new_bid.accept_bid(self.segment_hour)
#                     self.bids.append(new_bid)
#                     capacity_undercut = 0
#                 elif capacity_undercut > most_expensive_bid_capacity:
#                     del self.bids[-1]
#                     capacity_undercut = capacity_undercut - most_expensive_bid_capacity
#                     new_bid.accept_bid(self.segment_hour)
#                     self.bids.append(new_bid)
#                 elif capacity_undercut < most_expensive_bid_capacity:
#                     self.bids[-1].capacity_bid = self.bids[-1].capacity_bid - capacity_undercut
#                     new_bid.partially_accept_bid(self.segment_hour, capacity_undercut)
#                     self.bids.append(new_bid)
#                     capacity_undercut = 0
#                 else:
#                     new_bid.reject_bid(segment_hour=self.segment_hour)
#                     capacity_undercut = 0
#
#             # if self.capacity_required > 0:
#             #     self.bids.append(Bid(None, None, self.segment_hour, 0, elecsim.scenario.scenario_data.lost_load, self.model.year_number))
#
#         return self.bids
#
#
# class NoHistoryBidResponder(BidResponder):
#
#     def __init__(self, bids, capacity_required, segment_hour):
#         super().__init__(bids=bids, capacity_required=capacity_required, segment_hour=segment_hour)
#
#     def get_accepted_bids(self):
#         accepted_bids = [bid for bid in (self.classify_bids(bid) for bid in self.bids) if bid is not None]
#         return accepted_bids
#
#     def classify_bids(self, bid):
#         if self.capacity_required > 0 and self.capacity_required > bid.capacity_bid:
#             bid.accept_bid(self.segment_hour)
#             self.capacity_required -= bid.capacity_bid
#             # accepted_bids.append(bid)
#             logger.debug(
#                 'bid ACCEPTED: price: {}, year: {}, capacity required: {}, capacity: {}, capacity_bid: {}, type: {}, name {}'.format(
#                     bid.price_per_mwh, bid.plant.construction_year, self.capacity_required, bid.plant.capacity_mw,
#                     bid.capacity_bid, bid.plant.plant_type, bid.plant.name))
#             return bid
#         elif bid.capacity_bid > self.capacity_required > 0:
#             bid.partially_accept_bid(self.segment_hour, self.capacity_required)
#             self.capacity_required = 0
#             # accepted_bids.append(bid)
#             logger.debug(
#                 'bid PARTIALLY ACCEPTED: price: {}, year: {}, capacity required: {}, capacity: {}, capacity_bid: {}, type: {}, name {}'.format(
#                     bid.price_per_mwh, bid.plant.construction_year, self.capacity_required, bid.plant.capacity_mw,
#                     bid.capacity_bid, bid.plant.plant_type, bid.plant.name))
#             return bid
#         else:
#             bid.reject_bid(segment_hour=self.segment_hour)
#             logger.debug(
#                 'bid REJECTED: price: {}, year: {}, capacity required: {}, capacity: {}, capacity_bid: {}, type: {}, name {}'.format(
#                     bid.price_per_mwh, bid.plant.construction_year, self.capacity_required, bid.plant.capacity_mw,
#                     bid.capacity_bid, bid.plant.plant_type, bid.plant.name))
#             return None
#         # return capacity_required
#
#
#
#



"""fuel.py: Class which specifies the properties of a fuel"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "Alexander@Kell.es"


class Fuel():

    def __init__(self, fuel_type, fuel_price, energy_density, co2_density, mwh_to_co2e_conversion_factor=0.0):
        """
        Constructor which defines the characteristics of a plant_type
        :param fuel_type: Type of plant_type which is created
        :param fuel_price: List containing price of plant_type per year
        :param energy_density: Energy density of plant_type (GJ per tonne)
        :param co2_density: Carbon density of plant_type (tonnes of CO2 per tonne of gas burnt)
        :param mwh_to_co2e_conversion_factor: Conversion factor to convert MWh to CO2e emissions
        """

        self.fuel_type = fuel_type
        self.fuel_price = fuel_price
        self.energy_density = energy_density
        self.co2_density = co2_density
        self.mwh_to_co2e_conversion_factor = mwh_to_co2e_conversion_factor




from gym.envs.registration import register

"""
File name: __init__.py
Date created: 19/02/2019
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"

register(
    id='MyEnv-v0',
    entry_point='elecsim.reinforcement_learning.gym_elecsim.gym_elecsim.envs:WorldEnvironment',
)



import numpy as np
import numpy.random as ra
"""
File name: inverse_transform_sampling
Date created: 10/01/2019
Feature: # Generate random numbers from a custom distribution
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"


def sample_from_custom_distribution(count, division, n_samples):
    prob = [c/sum(count) for c in count]
    prob_sum = np.cumsum(prob)
    N=n_samples
    R=ra.uniform(0, 1, N)

    prob_sum = np.array(prob_sum)
    division = np.array(division)

    generate_points = [division[np.argwhere(prob_sum == min(prob_sum[(prob_sum - r) > 0]))][0][0] for r in R]
    if n_samples > 1:
        return generate_points
    elif n_samples == 1:
        return generate_points[0]



import logging
import warnings

from scipy.optimize import curve_fit, OptimizeWarning

from elecsim.data_manipulation.data_modifications.linear_regression import linear_regression

warnings.filterwarnings('error', category=OptimizeWarning)
import numpy as np

from elecsim.plants.plant_type.fuel_plant import FuelPlant

# import elecsim.scen_error.scenario_data as scenario
import elecsim.scenario.scenario_data

# from elecsim.scen_error.scenario_data import years_for_agents_to_predict_forward
logger = logging.getLogger(__name__)

"""
File name: current_market_data
Date created: 29/12/2018
Feature: # Data which collates market information for generator companies when making decisions about investments.
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"


class LatestMarketData:

    def __init__(self, model):
        self.model = model
        self.demand = self.model.demand

    def get_predicted_marginal_cost(self, power_plant, look_back_years):

        variable_o_m_cost = power_plant.variable_o_and_m_per_mwh

        if isinstance(power_plant, FuelPlant):

            co2_price = self.agent_forecast_value("co2", look_back_years, elecsim.scenario.scenario_data.years_for_agents_to_predict_forward)
            fuel_price = self.agent_forecast_value(power_plant.fuel.fuel_type, look_back_years, elecsim.scenario.scenario_data.years_for_agents_to_predict_forward)

            co2_cost = power_plant.fuel.mwh_to_co2e_conversion_factor * (1 / power_plant.efficiency) * co2_price
            fuel_cost = fuel_price/power_plant.efficiency
            short_run_marginal_cost = variable_o_m_cost + co2_cost + fuel_cost
        else:
            short_run_marginal_cost = variable_o_m_cost

        return short_run_marginal_cost

    def agent_forecast_value(self, value_required, years_to_look_back, years_to_look_forward=None, demand_linear=False):
        years_for_regression = list(range(self.model.years_from_start-years_to_look_back-1, self.model.years_from_start-1))
        variable_data = self._get_variable_data(value_required)

        regression = self._get_yearly_change_for_regression(variable_data, years_for_regression)
        regression_tuple = tuple(regression)
        if value_required != "demand" or demand_linear:
            next_value = linear_regression(regression_tuple, years_to_look_back, years_to_look_forward)
        else:
            next_value = self.fit_exponential_function(regression, years_to_look_back, years_to_look_forward)
        return next_value

    # def agent_forecast_demand(self, years_to_look_back, years_to_predict_forward):
    #     years_for_regression = list(range(self.model.step_number-years_to_look_back-1, self.model.step_number-1))
    #     scenario.yearly_demand_change

    @staticmethod
    def _get_variable_data(values_required):
        try:
            values_required = values_required.lower()
        except:
            raise ValueError("Price required must be a string, not a {}".format(type(values_required)))
        if values_required == "coal":
            return elecsim.scenario.scenario_data.coal_price
        elif values_required == "gas":
            return elecsim.scenario.scenario_data.gas_price
        elif values_required == "uranium":
            return elecsim.scenario.scenario_data.uranium_price
        elif values_required == "biomass_wood" or values_required == "woodchip":
            return elecsim.scenario.scenario_data.woodchip_price
        elif values_required == "biomass_poultry_litter" or values_required == "poultry_litter":
            return elecsim.scenario.scenario_data.poultry_litter_price
        elif values_required == "oil":
            return elecsim.scenario.scenario_data.oil_price
        elif values_required == "diesel" or values_required == "gas oil":
            return elecsim.scenario.scenario_data.diesel_price
        elif values_required == "straw":
            return elecsim.scenario.scenario_data.straw_price
        elif values_required == "meat":
            return elecsim.scenario.scenario_data.meat_price
        elif values_required == "waste_post_2000":
            return elecsim.scenario.scenario_data.waste_price_post_2000
        elif values_required == "waste_pre_2000":
            return elecsim.scenario.scenario_data.waste_price_pre_2000
        elif values_required == "co2":
            return elecsim.scenario.scenario_data.carbon_price_scenario
        elif values_required == "demand":
            return elecsim.scenario.scenario_data.yearly_demand_change
        else:
            raise ValueError("Could not find {}".format(values_required))


    def _get_yearly_change_for_regression(self, variable_data, years_for_regression):

        regression = [variable_data[i] if i > 0 else variable_data[0] for i in years_for_regression]
        return regression

    def fit_exponential_function(self, regression, years_to_look_back, years_to_look_forward):
        def exponential_func(x, a, b, c):
            return a * np.exp(b * x) + c

        regression = np.array(regression)

        x = np.array(list(range(1,years_to_look_back+1)))
        y = exponential_func(regression, 2.5, 1.3, 0.5)
        try:
            popt, pcov = curve_fit(exponential_func, x, regression)
        except (Warning, RuntimeError):
            return self.agent_forecast_value("demand", years_to_look_back, years_to_look_forward, demand_linear=True)

        return exponential_func(np.array(years_to_look_forward), *popt)

    def get_RL_investment_observations(self):

        years_for_regression = list(range(self.model.years_from_start-5-1, self.model.years_from_start-1))

        historical_co2 = self._get_yearly_change_for_regression(self._get_variable_data("co2"), years_for_regression)
        historical_coal = self._get_yearly_change_for_regression(self._get_variable_data("coal"), years_for_regression)
        historical_gas = self._get_yearly_change_for_regression(self._get_variable_data("gas"), years_for_regression)
        historical_uraniaum = self._get_yearly_change_for_regression(self._get_variable_data("uranium"), years_for_regression)
        historical_diesel = self._get_yearly_change_for_regression(self._get_variable_data("diesel"), years_for_regression)
        historical_demand = self._get_yearly_change_for_regression(self._get_variable_data("demand"), years_for_regression)

        # historical_coal = self._get_variable_data("coal")
        # historical_gas = self._get_variable_data("gas")
        # historical_uraniaum = self._get_variable_data("uranium")
        # historical_diesel = self._get_variable_data("diesel")
        # historical_demand = self._get_variable_data("demand")

        obs = {genco.name: np.hstack([historical_co2, historical_coal, historical_gas, historical_uraniaum, historical_diesel, historical_demand, genco.money]).reshape(31,) for genco in self.model.get_gencos()}

        return obs








from functools import lru_cache

import elecsim.scenario.scenario_data
from elecsim.constants import GJ_TO_MW
from elecsim.plants.fuel.fuel import Fuel

"""
File name: fuel_registry
Date created: 29/11/2018
Feature: #Enter feature description here
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"


@lru_cache()
def fuel_registry(fuel_type, fuel_price=None, energy_density=None, co2_density=None, mwh_to_co2e_conversion_factor=None):
    """
    Method which creates a Fuel object. Presets are defined for different fuels, and new fuels can be added
    based upon their name, carbon density and energy density
    :param fuel_type: Name of plant_type. Preset fuels are "Gas", "Coal", "Biomass_wood",
        "Biomass_poultry_litter", "Oil", "Diesel", "Straw", "Meat"
    :param fuel_price: Average price of fuel per year
    :param energy_density: Energy density of fuel (MWh per tonne)
    :param co2_density: Carbon density of fuel (tonnes of CO2 per tonne of fuel burnt)
    :param mwh_to_co2e_conversion_factor: Conversion factor to convert MWh to CO2e emissions. Source: Department for business, energy and industrial strategy: UK Government GHG Conversion Factors for Company Reporting,
    :return: Returns a Fuel object with characteristics of the plant_type in question
    """

    fuel_prices = elecsim.scenario.scenario_data.fuel_prices

    if energy_density is None and co2_density is None:
        if fuel_type == "Gas":
            return Fuel(fuel_type, fuel_prices[fuel_prices.Fuel == fuel_type], 35.8*GJ_TO_MW, 0.00203, 0.18396)
        if fuel_type == "Coal":
            return Fuel(fuel_type, fuel_prices[fuel_prices.Fuel == fuel_type], 24.1*GJ_TO_MW, 0.002231398, 0.31112)
        if fuel_type == "Uranium" or "Nuclear":
            return Fuel(fuel_type, fuel_prices[fuel_prices.Fuel == fuel_type], 1296000*GJ_TO_MW, 0, 0)
        if fuel_type == "Wind" or "Wind (offshore)":
            return Fuel(fuel_type, 0, 0, 0, 0)
        if fuel_type == "Solar":
            return Fuel(fuel_type, 0, 0, 0, 0)
        if fuel_type == "Hydro":
            return Fuel(fuel_type, 0, 0, 0, 0)
        if fuel_type == "Biomass_wood":
            return Fuel(fuel_type, fuel_prices[fuel_prices.Fuel == 'Woodchip'], 19.0*GJ_TO_MW, 0, 0.01506)
        if fuel_type == "Biomass_poultry_litter":
            return Fuel(fuel_type, fuel_prices[fuel_prices.Fuel == fuel_type], 7.9*GJ_TO_MW, 0, 0)
        if fuel_type == "Oil":
            return Fuel(fuel_type, fuel_prices[fuel_prices.Fuel == fuel_type], 40.7*GJ_TO_MW, 0.003200345, 0.26831)
        if fuel_type == "Diesel" or "Gas oil":
            return Fuel(fuel_type, fuel_prices[fuel_prices.Fuel == fuel_type], 42.6*GJ_TO_MW, 0.00319, 0.27652)
        if fuel_type == "Straw":
            return Fuel(fuel_type, fuel_prices[fuel_prices.Fuel == fuel_type], 13.1*GJ_TO_MW, 0, 0.01314)
        if fuel_type == "Meat":
            return Fuel(fuel_type, fuel_prices[fuel_prices.Fuel == fuel_type], 16.2*GJ_TO_MW, 0, 0)
        if fuel_type == "Waste_post_2000":
            return Fuel(fuel_type, fuel_prices[fuel_prices.Fuel == fuel_type], 13*GJ_TO_MW, 0.00033777, 0.001215972) # Taken average carbon emissions per tonne of mixture, and converted to conversion factor using tonne of CO2 emitted per tonne of waste. CO2 per tonne of waste taken from: Energy recovery for residual waste, a carbon based modelling approach. 11918_WR1910Energyrecoveryforresidualwaste-Acarbonbasedmodellingapporach.pdf
        if fuel_type == "Waste_pre_2000":
            return Fuel(fuel_type, fuel_prices[fuel_prices.Fuel == fuel_type], 13*GJ_TO_MW, 0.00033777, 0.001215972)
        else:
            raise ValueError("Must provide energy and carbon densities as plant_type is not preset.")
    elif co2_density is None:
        if energy_density > 0:
            return Fuel(fuel_type, fuel_price, energy_density,0)
        else:
            raise ValueError("Energy Density must be greater than 0")
    else:
        if energy_density > 0 & co2_density > 0:
            return Fuel(fuel_type,fuel_price, energy_density,co2_density)
        else:
            raise ValueError("Both energy density and co2 density must be larger than 0")



    # if energy_density is None and co2_density is None:
    #     if fuel_type == "Gas":
    #         return Fuel(fuel_type, scenario.gas_price, 35.8, 2.03)
    #     if fuel_type == "Coal":
    #         return Fuel(fuel_type, scenario.coal_price, 24.1, 2.23)
    #     if fuel_type == "Uranium" or "Nuclear":
    #         return Fuel(fuel_type, scenario.uranium_price, 1296000, 0)
    #     if fuel_type == "Wind" or "Wind (offshore)":
    #         return Fuel(fuel_type, 0, 0, 0)
    #     if fuel_type == "Solar":
    #         return Fuel(fuel_type, 0, 0, 0)
    #     if fuel_type == "Hydro":
    #         return Fuel(fuel_type, 0, 0, 0)
    #     if fuel_type == "Biomass_wood":
    #         return Fuel(fuel_type, scenario.woodchip_price, 19.0, 0)
    #     if fuel_type == "Biomass_poultry_litter":
    #         return Fuel(fuel_type, scenario.poultry_litter_price, 7.9, 0)
    #     if fuel_type == "Oil":
    #         return Fuel(fuel_type, scenario.oil_price, 40.7, 3.2)
    #     if fuel_type == "Diesel" or "Gas oil":
    #         return Fuel(fuel_type, scenario.diesel_price, 42.6, 3.19)
    #     if fuel_type == "Straw":
    #         return Fuel(fuel_type, scenario.straw_price, 13.1, 0)
    #     if fuel_type == "Meat":
    #         return Fuel(fuel_type, scenario.meat_price, 16.2, 0)
    #     if fuel_type == "Waste_post_2000":
    #         return Fuel(fuel_type, scenario.waste_price_post_2000, 13, 0)
    #     if fuel_type == "Waste_pre_2000":
    #         return Fuel(fuel_type, scenario.waste_price_pre_2000, 13, 0)
    #     else:
    #         raise ValueError("Must provide energy and carbon densities as plant_type is not preset.")
    # elif co2_density is None:
    #     if energy_density > 0:
    #         return Fuel(fuel_type, fuel_price, energy_density,0)
    #     else:
    #         raise ValueError("Energy Density must be greater than 0")
    # else:
    #     if energy_density > 0 & co2_density > 0:
    #         return Fuel(fuel_type,fuel_price, energy_density,co2_density)
    #     else:
    #         raise ValueError("Both energy density and co2 density must be larger than 0")


def plant_type_to_fuel(plant_type, construction_year=None):
    """
    Function which infers fuel type from plant plant_type.
    :return: String detailing fuel type
    """
    if plant_type == "CCGT":
        return "Gas"
    elif plant_type == "Recip_gas":
        return "Gas"
    elif plant_type == "Meat":
        return "Meat"
    elif plant_type == "Wind (offshore)":
        return None
    elif plant_type == "Offshore":
        return None
    elif plant_type == "Biomass_poultry_litter":
        return "Poultry_litter"
    elif plant_type == "Biomass_meat":
        return "Meat"
    elif plant_type == "Biomass_straw":
        return "Straw"
    elif plant_type == "Diesel":
        return "Diesel"
    elif plant_type == "Biomass_wood":
        return "Woodchip"
    elif plant_type == "Gas":
        return "Gas"
    elif plant_type == "Nuclear":
        return "Uranium"
    elif plant_type == "Coal":
        return "Coal"
    elif plant_type == "Solar":
        return None
    elif plant_type == "OCGT":
        return "Gas"
    elif plant_type == "Recip_diesel":
        return "Diesel"
    elif plant_type == "Gas oil":
        return "Diesel"
    elif plant_type == "EfW" and construction_year<=2000:
        return "Waste_pre_2000"
    elif plant_type == "EfW" and construction_year>2000:
        return "Waste_pre_2000"
    elif plant_type == "Pumped storage":
        return None
    elif plant_type == "Onshore":
        return None
    elif plant_type == "Offshore":
        return None
    elif plant_type == "Wind":
        return None
    elif plant_type == "PV":
        return None
    elif plant_type == "Pumped_storage":
        return None
    elif plant_type == "Hydro":
        return None

    else:
        raise ValueError("No fuel data for {}".format(plant_type))





"""
File name: ElecsimWorld
Date created: 19/02/2019
Feature: #Enter feature description here
"""
from gym.spaces import Box
import numpy as np

from elecsim.model.world import World
from elecsim.scenario.scenario_data import lost_load

import logging

logger = logging.getLogger(__name__)

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"

import gym
from gym import error, spaces, utils
from gym.utils import seeding


class WorldEnvironment(gym.Env):

    # def __init__(self, scenario_file=None, max_number_of_steps=32, data_folder="reinforcement_learning"):
    def __init__(self, config):
        print("trying to init")
        self.step_number_env = 0
        self.action = None
        self.config = config
        self.max_number_of_steps = config['max_number_of_steps']
        self.world = World(initialization_year=2018, scenario_file=config['scenario_file'], data_folder=config['data_folder'], number_of_steps=self.max_number_of_steps)
        self.action_space = Box(
            0.0, 250.0, shape=(1, ), dtype=np.float32)

        self.observation_space = Box(
            -10000.0, 0.0, shape=(2, ), dtype=np.float32)

    def reset(self):
        self.step_number_env=0
        self.world = World(initialization_year=2018, scenario_file=self.config['scenario_file'], data_folder=self.config['data_folder'], number_of_steps=self.max_number_of_steps)
        self.world.step_number -= 1
        resultant_observation = self.world.step(17)
        mean_electricity_price = resultant_observation[0]
        carbon_emitted = resultant_observation[1]
        obs = np.array([mean_electricity_price, carbon_emitted]) 
        # obs = -abs(mean_electricity_price)-abs(carbon_emitted)

        # obs = np.random.uniform(low=-100, high=1, size=(2,))
        obs = np.array(obs)
        # return np.expand_dims(obs, axis=0)
        return obs

    def step(self, action):
        self.action = action
        self.step_number_env += 1
        resultant_observation = self.world.step(action)
        mean_electricity_price = resultant_observation[0]
        carbon_emitted = resultant_observation[1]

        ob = -abs(mean_electricity_price)-abs(carbon_emitted)
        reward = ob
        done = self.world.step_number > self.max_number_of_steps
        print("step number: {}, action: {}, reward: {}, mean_electricity_price: {}, carbon_emitted: {}".format(self.world.step_number,action, reward, mean_electricity_price, carbon_emitted))

        ob = np.array([mean_electricity_price, carbon_emitted])

        return ob, reward, done, {}

    def render(self):

        return False



import logging
from functools import lru_cache

import numpy as np
import pandas as pd
import os, sys

from elecsim.market.electricity.market.power_exchange import PowerExchange
from elecsim.role.market.latest_market_data import LatestMarketData
import elecsim.scenario.scenario_data

logger = logging.getLogger(__name__)

"""
File name: predict_load_duration_prices
Date created: 11/01/2019
Feature: # Predict Load Duration Prices
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"


class PredictPriceDurationCurve:

    def __init__(self, model):
        self.model = model

    def predict_price_duration_curve(self, look_back_period):
        demand_change_predicted = LatestMarketData(self.model).agent_forecast_value("demand", look_back_period, elecsim.scenario.scenario_data.years_for_agents_to_predict_forward)

        # power_ex = PowerExchange(self.model)
        power_ex = self.model.PowerExchange
<<<<<<< HEAD
        power_ex.price_duration_curve = []
        if self.model.market_time_splices == 1:
            predicted_consumption = [cons * demand_change_predicted for cons in self.model.demand.segment_consumption]
            power_ex.tender_bids(self.model.demand.segment_hours, predicted_consumption, predict=True)
            # logger.info("bid length: {}".format(len(self.model.get_gencos()[0].plants[0].accepted_bids)))

            self.model.clear_all_bids()

=======
        if self.model.market_time_splices == 1:
            predicted_consumption = [cons * demand_change_predicted for cons in self.model.demand.segment_consumption]
            power_ex.tender_bids(self.model.demand.segment_hours, predicted_consumption, predict=True)
>>>>>>> 5f3c373861c398621fed06ebb3fe989ceb1733a9
            predicted_price_duration_curve = power_ex.price_duration_curve
            predicted_price_duration_curve = estimate_lost_load_price(predicted_price_duration_curve)
        else:
            price_duration_curve = []
            # year_segment_hours, year_segment_consumption = self.model.demand.get_demand_for_year()
            year_segment_hours = self.model.demand.year_segment_hours
            year_segment_consumption = self.model.demand.year_segment_consumption

            # for segment_hours, segment_consumption in zip(year_segment_hours, year_segment_consumption):

            year_segment_consumption_predicted = [cons * demand_change_predicted for cons in year_segment_consumption]

            power_ex.tender_bids(year_segment_hours, year_segment_consumption_predicted, predict=True)
<<<<<<< HEAD
            self.model.clear_all_bids()

=======
>>>>>>> 5f3c373861c398621fed06ebb3fe989ceb1733a9
            predicted_price_duration_curve = power_ex.price_duration_curve

            predicted_price_duration_curve = estimate_lost_load_price(predicted_price_duration_curve)

            # predicted_price_duration_curve.segment_hour = predicted_price_duration_curve.segment_hour.cumsum()
            # price_duration_curve.append(predicted_price_duration_curve_day)
            # predicted_price_duration_curve = pd.concat(price_duration_curve)
            # predicted_price_duration_curve.sort_values("segment_demand", ascending=False, inplace=True)

        return predicted_price_duration_curve


def estimate_lost_load_price(predicted_price_duration_curve):
    if elecsim.scenario.scenario_data.lost_load_price_predictor:
        if all(predicted_price_duration_curve.accepted_price == elecsim.scenario.scenario_data.lost_load):
            return predicted_price_duration_curve
        if any(predicted_price_duration_curve.accepted_price == elecsim.scenario.scenario_data.lost_load):
            predicted_price_duration_curve.accepted_price = predicted_price_duration_curve.accepted_price.replace(elecsim.scenario.scenario_data.lost_load,np.nan)
            # predicted_price_duration_curve.accepted_price.interpolate(method="polynomial", order=1)
            if predicted_price_duration_curve.accepted_price.count() > 1:
                predicted_price_duration_curve_training = predicted_price_duration_curve.dropna()

                p = np.poly1d(np.polyfit(predicted_price_duration_curve_training.segment_demand, predicted_price_duration_curve_training.accepted_price,1))
                extrapolated = p(predicted_price_duration_curve.loc[np.isnan(predicted_price_duration_curve.accepted_price), 'segment_demand'])
                logger.debug("extrapolated: {}".format(extrapolated))
                predicted_price_duration_curve.loc[np.isnan(predicted_price_duration_curve.accepted_price), "accepted_price"] = extrapolated
                return predicted_price_duration_curve
            elif predicted_price_duration_curve.accepted_price.count() == 1:
                predicted_price_duration_curve.accepted_price = predicted_price_duration_curve.accepted_price.fillna(predicted_price_duration_curve.accepted_price.mean())
                return predicted_price_duration_curve
        else:
            return predicted_price_duration_curve
    else:
        return predicted_price_duration_curve


# @lru_cache(1024)
def   get_price_duration_curve(model, look_back_period):
    predicted_price_duration_curve = PredictPriceDurationCurve(model).predict_price_duration_curve(look_back_period=look_back_period)
    return predicted_price_duration_curve




from functools import lru_cache
import logging

from elecsim.plants.fuel.fuel_registry.fuel_registry import fuel_registry, plant_type_to_fuel
from elecsim.plants.plant_type.power_plant import PowerPlant
from elecsim.role.plants.costs.fuel_plant_cost_calculations import FuelPlantCostCalculations



logger = logging.getLogger(__name__)

""" fuel_plant.py: Child class of power plant which contains functions for a power plant which consumes fuel.
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "Alexander@Kell.es"


class FuelPlant(PowerPlant):

    def __init__(self, name, plant_type, capacity_mw, construction_year, average_load_factor, efficiency, pre_dev_period, construction_period, operating_period, pre_dev_spend_years, construction_spend_years, pre_dev_cost_per_mw, construction_cost_per_mw, infrastructure, fixed_o_and_m_per_mw, variable_o_and_m_per_mwh, insurance_cost_per_mw, connection_cost_per_mw):
        """
        Initialisation of plant_type power plant object.
        :param efficiency: Efficiency of power plant at converting plant_type energy into electrical energy.
        """
        super().__init__(name=name, plant_type=plant_type, capacity_mw=capacity_mw, construction_year=construction_year, average_load_factor=average_load_factor, pre_dev_period=pre_dev_period, construction_period=construction_period, operating_period=operating_period, pre_dev_spend_years=pre_dev_spend_years, construction_spend_years=construction_spend_years, pre_dev_cost_per_mw=pre_dev_cost_per_mw, construction_cost_per_mw=construction_cost_per_mw, infrastructure=infrastructure, fixed_o_and_m_per_mw=fixed_o_and_m_per_mw, variable_o_and_m_per_mwh=variable_o_and_m_per_mwh, insurance_cost_per_mw=insurance_cost_per_mw, connection_cost_per_mw=connection_cost_per_mw)
        self.efficiency = efficiency
        # Finds fuel type of power plant eg. CCGT power plant plant_type returns gas.
        fuel_string = plant_type_to_fuel(plant_type, self.construction_year)
        # Fuel object, containing information on fuel.
        self.fuel = fuel_registry(fuel_string)

        if plant_type in ['Coal', 'Nuclear']:
            self.min_running = 5000
        else:
            self.min_running = 0


    def calculate_lcoe(self, discount_rate):
        plant_cost_calculations = FuelPlantCostCalculations(plant_type=self.plant_type, capacity_mw=self.capacity_mw, construction_year=self.construction_year, average_load_factor=self.average_load_factor, efficiency=self.efficiency, pre_dev_period=self.pre_dev_period, construction_period=self.construction_period, operating_period=self.operating_period, pre_dev_spend_years=self.pre_dev_spend_years, construction_spend_years=self.construction_spend_years, pre_dev_cost_per_mw=self.pre_dev_cost_per_mw, construction_cost_per_mw=self.construction_cost_per_mw, infrastructure=self.infrastructure, fixed_o_and_m_per_mw=self.fixed_o_and_m_per_mw, variable_o_and_m_per_mwh=self.variable_o_and_m_per_mwh, insurance_cost_per_mw=self.insurance_cost_per_mw, connection_cost_per_mw=self.connection_cost_per_mw)
        lcoe = plant_cost_calculations.calculate_lcoe(discount_rate)
        return lcoe


    def short_run_marginal_cost(self, model, genco, fuel_price = None, co2_price = None):
        plant_cost_calculations = FuelPlantCostCalculations(plant_type=self.plant_type, capacity_mw=self.capacity_mw, construction_year=self.construction_year, average_load_factor=self.average_load_factor, efficiency=self.efficiency, pre_dev_period=self.pre_dev_period, construction_period=self.construction_period, operating_period=self.operating_period, pre_dev_spend_years=self.pre_dev_spend_years, construction_spend_years=self.construction_spend_years, pre_dev_cost_per_mw=self.pre_dev_cost_per_mw, construction_cost_per_mw=self.construction_cost_per_mw, infrastructure=self.infrastructure, fixed_o_and_m_per_mw=self.fixed_o_and_m_per_mw, variable_o_and_m_per_mwh=self.variable_o_and_m_per_mwh, insurance_cost_per_mw=self.insurance_cost_per_mw, connection_cost_per_mw=self.connection_cost_per_mw)
        marginal_cost = plant_cost_calculations.calculate_short_run_marginal_cost(model, genco, fuel_price, co2_price)

        return marginal_cost


    def __repr__(self):
        return 'FuelPlant({}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {})'.format(self.name, self.plant_type, self.capacity_mw, self.construction_year, self.average_load_factor, self.pre_dev_period, self.construction_period, self.operating_period, self.pre_dev_spend_years, self.construction_spend_years, self.pre_dev_cost_per_mw, self.construction_cost_per_mw, self.infrastructure, self.fixed_o_and_m_per_mw, self.variable_o_and_m_per_mwh, self.insurance_cost_per_mw, self.connection_cost_per_mw)





import logging
import math
from random import gauss
import numpy as np
from functools import lru_cache
from mesa import Agent
# from linetimer import CodeTimer

from elecsim.market.electricity.bid import Bid
from elecsim.plants.availability_factors.availability_factor_calculations import get_availability_factor
from elecsim.plants.fuel.capacity_factor.capacity_factor_calculations import get_capacity_factor
from elecsim.plants.plant_costs.estimate_costs.estimate_costs import create_power_plant, create_power_plant_group
from elecsim.plants.plant_type.fuel_plant import FuelPlant
from elecsim.role.investment.calculate_npv import get_most_profitable_plants_by_npv
from elecsim.role.investment.calculate_npv import select_yearly_payback_payment_for_year
from elecsim.role.market.latest_market_data import LatestMarketData
# from elecsim.scen_error.scenario_data import bid_mark_up, pv_availability, offshore_availability, onshore_availability, non_fuel_plant_availability
# from elecsim.scen_error.scenario_data import nuclear_wacc, non_nuclear_wacc, upfront_investment_costs, years_for_agents_to_predict_forward

import elecsim.scenario.scenario_data

logger = logging.getLogger(__name__)

"""gen_co.py: Agent which represents a generation company"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "Alexander@Kell.es"


class GenCo(Agent):
    def __init__(self, unique_id, model, name, difference_in_discount_rate, look_back_period, plants=None,
                 money=5000000):
        """
        Agent which defines a generating company
        :param unique_id: Unique ID for the generating company
        :param model:  Model which defines the world that the agent lives in
        :param name: Name of generating company
        :param plants: Plants which the generating company is initialised with
        :param money: Money which the agent is initialised with
        """
        super().__init__(unique_id, model)
        if plants is None:
            plants = []

        self.name = name
        self.plants = plants
        self.money = money

        self.difference_in_discount_rate = difference_in_discount_rate
        self.look_back_period = look_back_period

        self.gas_price_modifier = 0
        self.coal_price_modifier = 0


    def step(self):
        logger.debug("Stepping generation company: {}".format(self.name))
        logger.debug("Amount of money: {}".format(self.money))
        self.delete_old_bids()
        if elecsim.scenario.scenario_data.investment_mechanism != "RL" and self.model.step_number % self.model.market_time_splices == 0 and self.model.step_number != 0 and self.model.continue_investing < 3:
        # if elecsim.scenario.scenario_data.investment_mechanism != "RL" and self.model.step_number % self.model.market_time_splices == 0 and self.model.continue_investing < 3:
            continue_investing = self.invest()
            if continue_investing == 1:
                self.model.continue_investing += continue_investing
                logger.debug(self.model.continue_investing)
            else:
                self.model.continue_investing = 0

        # self.reset_contracts()
        self.purchase_fuel()

    # def calculate_non_fuel_bids(self, segment_hour):
    #     bids = []
    #     renewable_plants = [renewable_plant for renewable_plant in self.plants if isinstance(renewable_plant, NonFuelPlant)]
    #     for plant in renewable_plants:
    #         marked_up_price = plant.short_run_marginal_cost(self.model)*1.1
    #         bids.append(Bid(self, plant, segment_hour, plant.capacity_mw, marked_up_price))
    #     return bids

    def calculate_bids(self, segment_hour, predict=False):
        """
        Function to generate the bids for each of the power plants owned by the generating company.
        The bids submitted are the fixed costs divided by lifetime of plant plus yearly variable costs plus a 10% margin
        :param segment_hour: Number of hours in which the current segment is required
        :param segment_demand: Electricity consumption required for the specified number of hours
        :return: Bids returned for the available plants at the specified segment hour
        """

        bids = []

        for plant in self.plants:
            bid = self.create_bid(plant, predict, segment_hour, self.model.step_number)
            if bid:
                bids.append(bid)
        return bids

    @lru_cache(100000)
    def create_bid(self, plant, predict, segment_hour, step_number):
        future_plant_operating = False
        if predict is True:
            YEARS_IN_FUTURE_TO_PREDICT_SUPPLY = elecsim.scenario.scenario_data.years_for_agents_to_predict_forward
            future_plant_operating = plant.check_if_operating_in_certain_year(self.model.year_number, YEARS_IN_FUTURE_TO_PREDICT_SUPPLY)

            if isinstance(plant, FuelPlant):
                if future_plant_operating:
                    price = self._get_predicted_short_run_marginal_cost(plant)
                else:
                #     price = plant.short_run_marginal_cost(self.model, self)
                    return None
            else:
                price = plant.short_run_marginal_cost(self.model, self)
        else:
            price = plant.short_run_marginal_cost(self.model, self)
        marked_up_price = price * elecsim.scenario.scenario_data.bid_mark_up
        if plant.is_operating or future_plant_operating:
            if plant.plant_type in ['Offshore', 'Onshore', 'PV', 'Hydro', "Nuclear"]:
                capacity_factor, availability = _create_bid_for_capacity_factor_available_plants(
                    self.model.market_time_splices, plant, segment_hour)
                # logger.info(_create_bid_for_capacity_factor_available_plants.cache_info())
                return Bid(self, plant, segment_hour, capacity_factor * availability * plant.capacity_mw, marked_up_price,
                        self.model.year_number)
            elif isinstance(plant, FuelPlant):
                if plant.capacity_fulfilled[segment_hour] < plant.capacity_mw:
                    return self._create_bid_for_fuel_plant(marked_up_price, plant, segment_hour)
            elif plant.plant_type != 'Hydro_Store':
                return Bid(self, plant, segment_hour,
                        elecsim.scenario.scenario_data.non_fuel_plant_availability * plant.capacity_mw, marked_up_price,
                        self.model.year_number)

    def _create_bid_for_fuel_plant(self, marked_up_price, plant, segment_hour):
        availability_factor = get_availability_factor(plant_type=plant.plant_type,
                                                      construction_year=plant.construction_year)
        capacity_to_bid = availability_factor * plant.capacity_mw
        # logger.info("plant_type: {}, construction_year: {}, capacity_to_bid: {}, plant capacity: {}, availability factor: {}".format(plant.plant_type, plant.construction_year, capacity_to_bid, plant.capacity_mw, availability_factor))
        return Bid(self, plant, segment_hour, capacity_to_bid, marked_up_price, self.model.year_number)


    def _get_predicted_short_run_marginal_cost(self, plant):
        co2_price_predicted = self.forecast_attribute_price("co2")
        fuel_price_predicted = self.forecast_attribute_price(plant.fuel.fuel_type)
        price = plant.short_run_marginal_cost(self.model, self, fuel_price_predicted, co2_price_predicted)
        return price


    def invest(self):

        lowest_upfront_cost = 0
        down_payment = 0
        counter = 0
        total_capacity = 0
        number_of_plants_to_purchase = 1
        # while self.money > lowest_upfront_cost and total_capacity < 1500:
        while self.money > lowest_upfront_cost:
        # while number_of_plants_to_purchase > 0:

            counter += 1
            # if counter>3:
            #     break
            # potential_plant_data = npv_calculation.get_positive_npv_plants_list()
            potential_plant_data = get_most_profitable_plants_by_npv(self.model, self.difference_in_discount_rate,
                                                                     self.look_back_period)

            # logger.info("potential_plant_data: {}".format(potential_plant_data))
            if potential_plant_data:
                potential_plant_list = []
                for plant_data in potential_plant_data:
                    power_plant_trial = create_power_plant("invested_plant", self.model.year_number, plant_data[1], plant_data[0])
                    potential_plant_list.append(power_plant_trial)
                lowest_upfront_cost = min(plant.get_upfront_costs() * elecsim.scenario.scenario_data.upfront_investment_costs for plant in potential_plant_list)
            else:
                return 1

            for plant_data in potential_plant_list:

                # power_plant_trial = create_power_plant("invested_plant", self.model.year_number, plant_data[1], plant_data[0])
                power_plant_trial = create_power_plant("invested_plant", self.model.year_number, plant_data.plant_type, plant_data.capacity_mw)
                down_payment = power_plant_trial.get_upfront_costs() * elecsim.scenario.scenario_data.upfront_investment_costs
                number_of_plants_to_purchase = int(self.money/down_payment)
                if number_of_plants_to_purchase == 0:
                    continue

                capacity_to_purchase = number_of_plants_to_purchase * power_plant_trial.capacity_mw
                if capacity_to_purchase > 200000:
                    number_of_plants_to_purchase = int(200000/power_plant_trial.capacity_mw)

                power_plant_trial_group = create_power_plant_group(name="invested_plant_group", start_date=self.model.year_number, simplified_type=plant_data.plant_type, capacity=plant_data.capacity_mw, number_of_plants_to_purchase=number_of_plants_to_purchase)
                down_payment_of_plant_array = number_of_plants_to_purchase*down_payment
                # logger.info(create_power_plant.cache_info())
                if self.money > down_payment_of_plant_array and number_of_plants_to_purchase >= 1:
                    logger.info("investing in {}, company: {}, size: {}, number: {}, self.money: {}".format(power_plant_trial_group.plant_type, self.name, power_plant_trial_group.capacity_mw, number_of_plants_to_purchase, self.money))
                    self.plants.append(power_plant_trial_group)
                    self.money -= down_payment_of_plant_array
                    total_capacity += power_plant_trial_group.capacity_mw

                    break


    def invest_RL(self, action):
        plant_list = elecsim.scenario.scenario_data.potential_plants_to_invest

        plant_string_to_invest = plant_list[action.item(0)]
        plant = elecsim.scenario.scenario_data.modern_plant_costs[elecsim.scenario.scenario_data.modern_plant_costs.Plant_Type.str.contains(plant_string_to_invest)]

        plant_group = create_power_plant_group("plant_RL_invested", self.model.year_number, plant.Type.values[0], plant.Plant_Size.values[0], action.item(1))

        self.plants.append(plant_group)


    def dismantle_old_plants(self):
        """
        Remove plants that are past their lifetime agent from plant list
        """

        def get_running_plants(plants):
            for plant in plants:
                if plant.construction_year + plant.operating_period + plant.construction_period + plant.pre_dev_period >= self.model.year_number:
                    yield plant
                else:
                    logger.debug("Taking the plant '{}' out of service, year of construction: {}".format(plant.name, plant.construction_year))
                    for bid in self.model.PowerExchange.stored_ordered_bids:
                        if bid.plant == plant:
                            del bid

        plants_filtered = list(get_running_plants(self.plants))
        self.plants = plants_filtered

    # def pay_money(self):
    #     # self.money -=
    #     # fixed_costs = sum(plant for plant in self.plants)
    #     # expenditure = sum(bid for plant in self.plants for bid in plant.bids)
    #
    #     for plant in self.plants:
    #         pass

    def settle_accounts(self):
        # income = sum((bid.price_per_mwh * bid.segment_hours) for plant in self.plants for bid in plant.accepted_bids if
        #              bid.partly_accepted or bid.bid_accepted)
        #
        # short_run_marginal_expenditure = sum((bid.segment_hours * bid.capacity_bid * plant.short_run_marginal_cost(model=self.model, genco=self)
        #                                       for plant in self.plants for bid in plant.accepted_bids if bid.partly_accepted or bid.bid_accepted if plant.is_operating == True))
        net_income = 0
        income = 0
        short_run_marginal_expenditure = 0
        for plant in self.plants:
            previous_segment_hour = 0
            if self.model.market_time_splices == 1:
                bids_ordered = reversed(plant.accepted_bids)
            else:
                bids_ordered = plant.accepted_bids
            for bid in bids_ordered:
                if bid.partly_accepted or bid.bid_accepted:
                    income += bid.price_per_mwh * (bid.segment_hours - previous_segment_hour) * bid.capacity_bid
                    if plant.is_operating:
                        srmc = plant.short_run_marginal_cost(model=self.model, genco=self)
                        short_run_marginal_expenditure += (bid.segment_hours - previous_segment_hour) * bid.capacity_bid * srmc
                previous_segment_hour = bid.segment_hours

            percentage_of_time_of_year = self.model.demand.segment_hours[-1]/8760
            logger.debug("percentage_of_time_of_year: {}".format(percentage_of_time_of_year))

            interest = [elecsim.scenario.scenario_data.nuclear_wacc if plant.plant_type == "Nuclear" else elecsim.scenario.scenario_data.non_nuclear_wacc for plant in self.plants]

            fixed_variable_costs = sum((plant.fixed_o_and_m_per_mw * plant.capacity_mw) for plant in self.plants if plant.is_operating is True)

            capital_loan_expenditure = sum(select_yearly_payback_payment_for_year(plant, interest_rate + self.difference_in_discount_rate, elecsim.scenario.scenario_data.upfront_investment_costs, self.model) for plant, interest_rate in zip(self.plants, interest))

            cashflow = income - short_run_marginal_expenditure*percentage_of_time_of_year - fixed_variable_costs*percentage_of_time_of_year + capital_loan_expenditure*percentage_of_time_of_year
            net_income += cashflow
        # logger.debug("income: {}, outflow: {}".format(income, short_run_marginal_expenditure/percentage_of_time_of_year - fixed_variable_costs/percentage_of_time_of_year + capital_loan_expenditure/percentage_of_time_of_year))
        logger.debug("cashflow: {} for {}".format(net_income, self.name))
        if not math.isnan(net_income):
            self.money += net_income

    # def reset_contracts(self):
    #     """
    #     Function to reset the contracts of all plants
    #     :return: None
    #     """
    #     for plant in self.plants:
    #         plant.reset_plant_contract()

    def delete_old_bids(self):
        for plant in self.plants:
            plant.delete_old_plant_bids()

    def purchase_fuel(self):
        if any(plant.plant_type == "CCGT" for plant in self.plants):
            self.purchase_gas()
        if any(plant.plant_type == "Coal" for plant in self.plants):
            self.purchase_coal()

    def purchase_gas(self):
        self.gas_price_modifier = gauss(mu=-0.010028, sigma=0.634514)

    def purchase_coal(self):
        self.coal_price_modifier = gauss(mu=-0.042884, sigma=3.400331)

    def forecast_demand_change(self):
        latest_market_data = LatestMarketData(self.model)
        demand_change_predicted = latest_market_data.agent_forecast_value("demand", self.look_back_period, elecsim.scenario.scenario_data.years_for_agents_to_predict_forward)
        return demand_change_predicted

    def forecast_attribute_price(self, fuel_type):
        latest_market_data = LatestMarketData(self.model)
        predicted_fuel_price = latest_market_data.agent_forecast_value(fuel_type, self.look_back_period, elecsim.scenario.scenario_data.years_for_agents_to_predict_forward)
        return predicted_fuel_price


@lru_cache(64)
def get_renewable_availability(plant):
    if plant.plant_type == "Offshore":
        availability = elecsim.scenario.scenario_data.offshore_availability
    elif plant.plant_type == 'Onshore':
        availability = elecsim.scenario.scenario_data.onshore_availability
    elif plant.plant_type == "PV":
        availability = elecsim.scenario.scenario_data.pv_availability
    else:
        availability = elecsim.scenario.scenario_data.non_fuel_plant_availability
    return availability

@lru_cache(250000)
def _create_bid_for_capacity_factor_available_plants(market_time_splices, plant, segment_hour):
    capacity_factor = get_capacity_factor(market_time_splices, plant.plant_type, segment_hour)
    availability = get_renewable_availability(plant)
    return capacity_factor, availability



import logging
import math
from random import uniform
from functools import lru_cache

from scipy.interpolate import interp1d

import elecsim.scenario.scenario_data
from elecsim.data_manipulation.data_modifications.extrapolation_interpolate import ExtrapolateInterpolate
from elecsim.plants.fuel.fuel_registry.fuel_registry import plant_type_to_fuel

logger = logging.getLogger(__name__)

from elecsim.data_manipulation.data_modifications.value_estimations import closest_row



class PredictModernPlantParameters:
    def __init__(self, plant_type, capacity, start_year):
        """
        Class which provides calculations to provide costing data for power plants based on plant_type, capacity and start year
        :param plant_type (str): Type of plant_type that plant runs on
        :param capacity (int): Capacity of plant in MW
        :param start_year (int): Year that power plant begun construction
        """
        self.plant_type = plant_type
        self.capacity = float(capacity)
        self.start_year = float(start_year)

        # Import UK power plant cost data
        self.cost_data = elecsim.scenario.scenario_data.modern_plant_costs
        self.year_data = [int(sub.split(" _")[1]) for sub in self.cost_data.filter(regex="Connect_system_cost-Medium ").columns]
        # self.cost_data = self.cost_data[self.cost_data.Type == self.plant_type].sort_values('Plant_Size')
        self.cost_data = self.cost_data[self.cost_data.apply(lambda x: x['Type'] in self.plant_type, axis=1)].sort_values('Plant_Size')

    def parameter_estimation(self):
        """
        Function which estimates costs of power plant based on capacity, plant_type and start year. Use of linear interpolation
        for plants of capacity that fall within known range of costing variables. If plant capacity to be calculated
        falls out of range of known values, the highest or lowest capacity value is chosen.
        :return (:obj:`dict`:obj:`str`): Returns dictionary containing variables for PowerPlant cost
        """

        # Iterates through each type of plant cost to predict parameters.
        initial_stub_cost_parameters = ['Connect_system_cost-Medium _', 'Constr_cost-Medium _', 'Fixed_cost-Medium _',
                                        'Infra_cost-Medium _', 'Insurance_cost-Medium _', 'Pre_dev_cost-Medium _',
                                        'Var_cost-Medium _']

        condition = True
        while condition:
            full_cost_parameters = self._create_parameter_names(initial_stub_cost_parameters)

            parameters_of_plant = {
                self._change_columns(cost_variable_required): ExtrapolateInterpolate(self.cost_data['Plant_Size'],
                                                                                     self.cost_data[
                                                                                         cost_variable_required])(self.capacity)
                for cost_variable_required in full_cost_parameters}
            self.check_plant_exists(parameters_of_plant)

            if all(math.isnan(value) for value in parameters_of_plant.values()):
                self.start_year += 1
            else:
                condition = False

        durations = ['Pre_Dur', 'Operating_Period', 'Constr_Dur', 'Efficiency', 'Average_Load_Factor']
        durations_parameters = {self._change_columns(dur): self._estimate_non_interpolatable_parameters(dur) for dur in durations}
        # logger.info(self._estimate_non_interpolatable_parameters.cache_info())
        yearly_cost_spread = ['Constr', 'Pre']

        yearly_cost_perc = {self._change_columns(spread): self._payment_spread_estimator(spread) for spread in
                            yearly_cost_spread}

        parameters = {**parameters_of_plant, **durations_parameters, **yearly_cost_perc}
        parameters = self.check_pre_dev_spend(parameters)
        parameters['variable_o_and_m_per_mwh'] *= uniform(elecsim.scenario.scenario_data.o_and_m_multiplier[0], elecsim.scenario.scenario_data.o_and_m_multiplier[1])
        self._use_historical_efficiency_data(parameters)

        return parameters

    def _use_historical_efficiency_data(self, parameters):
        if self.plant_type in ['CCGT', 'Coal'] and self.start_year < 2018:
            fuel_used = plant_type_to_fuel(self.plant_type)
            historical_efficiency_measure = elecsim.scenario.scenario_data.historical_fuel_plant_efficiency[
                elecsim.scenario.scenario_data.historical_fuel_plant_efficiency.fuel_type == fuel_used]

            historical_efficiency_measure = historical_efficiency_measure.reset_index()
            efficiency = ExtrapolateInterpolate(historical_efficiency_measure.Year,
                                                historical_efficiency_measure.efficiency).min_max_extrapolate(
                self.start_year)
            parameters['efficiency'] = efficiency

    @staticmethod
    def check_pre_dev_spend(parameters):
        if not parameters['pre_dev_spend_years']:
            parameters['pre_dev_cost_per_mw'] = 0
        if not parameters['construction_spend_years']:
            parameters['construction_cost_per_mw'] = 0
        return parameters

    def check_plant_exists(self, parameters_of_plant):
        """
        Function which checks that there is data for specified power plant in the modern costs database.
        :param parameters_of_plant: Dictionary of plant which contains all of the values of the estimated power plant.
        """
        if all(value == 0 for value in parameters_of_plant.values()):
            raise ValueError("No cost data for power plant of type: " + self.plant_type)

    def _create_parameter_names(self, initial_stub_cost_parameters):
        """
        Function that chooses the names of the parameters to search the data file of modern plant costs. For instance,
        choose Connect_system_cost-Medium _2018 for plant built in 2018 and 2019.
        :param plant_costs: 
        :return:
        """


        if self.start_year in self.year_data:
            cost_parameter_variables = [cost_variable + str(int(self.start_year)) for cost_variable in
                                        initial_stub_cost_parameters]
        elif self.start_year > self.year_data[-1]:
            cost_parameter_variables = [cost_variable + str(self.year_data[-1]) for cost_variable in
                                        initial_stub_cost_parameters]
        elif self.year_data[0] < self.start_year < self.year_data[1]:
            cost_parameter_variables = [cost_variable + str(self.year_data[0]) for cost_variable in
                                        initial_stub_cost_parameters]
        elif self.year_data[1] < self.start_year < self.year_data[2]:
            cost_parameter_variables = [cost_variable + str(self.year_data[1]) for cost_variable in
                                        initial_stub_cost_parameters]
        elif self.start_year < self.year_data[0]:
            cost_parameter_variables = [cost_variable + str(self.year_data[0]) for cost_variable in
                                        initial_stub_cost_parameters]

        return cost_parameter_variables

    # @lru_cache(maxsize=10000)
    def _estimate_non_interpolatable_parameters(self, variable_wanted):
        """
        Estimates parameters time scale required for construction, pre-development and operating period.
        This is done by selecting the operating period, construction and pre-development of the closest sized
        power plant in data
        :param variable_wanted (str): Variable that is required to estimate
        :return (int): Returns estimated duration parameter in years.
        """
        column_required = self.cost_data[['Plant_Size', variable_wanted]].dropna()
        if min(column_required.Plant_Size) < self.capacity < max(column_required.Plant_Size):
            interp = interp1d(column_required.Plant_Size, column_required[variable_wanted], kind='nearest')
            return interp(self.capacity)
        elif self.capacity >= max(column_required.Plant_Size):
            column_required = column_required.reset_index()
            return column_required.iloc[-1][variable_wanted]
        elif self.capacity <= min(column_required.Plant_Size):
            column_required = column_required.reset_index()
            return column_required.iloc[0][variable_wanted]

    def _payment_spread_estimator(self, var_wanted):
        """
        Function which selects the spread of payments required for construction and pre-development. This is achieved
        by selecting the power plant of the closest size.
        :param var_wanted (str): Variable to estimate spread of payments.
        :return (:obj:`list` of :obj:`str`): Returns list of percentage of cost per year
        """
        # df_sort = self.cost_data.iloc[(self.cost_data['Plant_Size']-self.capacity).abs().argsort()[:1]]
        df_sort = closest_row(self.cost_data, "Plant_Size", self.capacity)
        df_sort = df_sort.filter(regex=var_wanted).filter(regex='^((?!Dur).)*$').filter(regex='^((?!-).)*$').dropna(
            axis=1).values.tolist()[0]

        return df_sort

    @staticmethod
    def _change_columns(column):
        """
        Function which converts variable names from UK power plant data, to PowerPlant instance variable names for easy
        conversion
        :param column (str): String name of variable from UK power plant data
        :return (str): Name in PowerPlant instance variable name format
        """
        if 'Connect_system_cost' in column:
            return 'connection_cost_per_mw'
        elif 'Constr_cost' in column:
            return 'construction_cost_per_mw'
        elif 'Fixed_cost' in column:
            return 'fixed_o_and_m_per_mw'
        elif 'Infra_cost' in column:
            return 'infrastructure'
        elif 'Insurance_cost' in column:
            return 'insurance_cost_per_mw'
        elif 'Pre_dev_cost' in column:
            return 'pre_dev_cost_per_mw'
        elif 'Var_cost' in column:
            return 'variable_o_and_m_per_mwh'
        elif 'Pre_Dur' in column:
            return 'pre_dev_period'
        elif 'Operating_Period' in column:
            return 'operating_period'
        elif 'Constr_Dur' in column:
            return 'construction_period'
        elif 'Constr' in column:
            return 'construction_spend_years'
        elif 'Pre' in column:
            return 'pre_dev_spend_years'
        elif 'Efficiency' in column:
            return 'efficiency'
        elif 'Average_Load_Factor' in column:
            return 'average_load_factor'
        else:
            raise ValueError('Plant cost data not found')



import numpy as np

"""
File name: curve_fitting
Date created: 10/07/2019
Feature: #Enter feature description here
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"


def logit(x, a, b, c, d, e):
    return -a*np.log(b*x/(c-d*x)) + e




"""power_plant.py: Class which represents a Power Plant"""

import logging
from abc import ABC, abstractmethod

logger = logging.getLogger(__name__)
__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "Alexander@Kell.es"


class PowerPlant(ABC):
    def __init__(self, name, plant_type, capacity_mw, construction_year, average_load_factor, pre_dev_period, construction_period, operating_period, pre_dev_spend_years, construction_spend_years, pre_dev_cost_per_mw, construction_cost_per_mw, infrastructure, fixed_o_and_m_per_mw, variable_o_and_m_per_mwh, insurance_cost_per_mw, connection_cost_per_mw):
        """
        PowerPlant class which are built and operated by generation companies
        :param name: Name of power plant
        :param plant_type: Type of power plant
        :param capacity_mw: Capacity of power plant (MW)
        :param construction_year: Year that the power plant was constructed
        :param average_load_factor: Average amount of time that power plant is used
        :param pre_dev_period: Amount of time construction spends in pre-development
        :param construction_period: Amount of time construction spends in construction
        :param construction_spend_years: Percentage of costs that are spread over each year of construction
        :param pre_dev_spend_years: Percentage of costs that are spread over each year of pre-development
        :param operating_period: How long the power plant remains operational
        :param pre_dev_cost_per_mw: Cost of pre-development per kW of capacity
        :param construction_cost_per_mw: Cost of construction per kW of capacity
        :param infrastructure: Infrastructure cost in GBP
        :param fixed_o_and_m_per_mw: Fixed operation and maintenance cost
        :param variable_o_and_m_per_mwh: Variable operation and maintenance cost
        :param insurance_cost_per_mw: Insurance cost
        :param connection_cost_per_mw: Connection and use of system cost
        """

        # Data from BEIS
        self.name = name

        self.plant_type = plant_type

        self.capacity_mw = capacity_mw

        self.construction_year = construction_year

        self.average_load_factor = average_load_factor

        self.pre_dev_period = pre_dev_period
        self.pre_dev_spend_years = pre_dev_spend_years

        self.construction_period = construction_period
        self.construction_spend_years = construction_spend_years

        self.pre_dev_cost_per_mw = pre_dev_cost_per_mw

        self.construction_cost_per_mw = construction_cost_per_mw

        self.operating_period = operating_period

        self.infrastructure = infrastructure

        self.fixed_o_and_m_per_mw = fixed_o_and_m_per_mw

        self.variable_o_and_m_per_mwh = variable_o_and_m_per_mwh

        self.insurance_cost_per_mw = insurance_cost_per_mw

        self.connection_cost_per_mw = connection_cost_per_mw

        self.is_operating = False


        #
        # self.min_running = min_running


        # Bids
        self.accepted_bids = []
        self.historical_bids = []

        self.capacity_fulfilled = {
            0.08: 0,
            460.75: 0,
            921.42: 0,
            1382.08: 0,
            1842.67: 0,
            2303.33: 0,
            2764: 0,
            3224.67: 0,
            3685.33: 0,
            4146: 0,
            4606.58: 0,
            5067.25: 0,
            5527.92: 0,
            5988.58: 0,
            6449.25: 0,
            6909.92: 0,
            7370.5: 0,
            7831.17: 0,
            8291.83: 0,
            8752.5: 0
        }

    # @property
    # def infrastructure(self):
    #     return self._infrastructure
    #
    # @infrastructure.setter
    # def infrastructure(self, value):
    #     self._infrastructure = value * constants.KW_TO_MW


    #     # Bids
    #     self.accepted_bids = []


    # def reset_plant_contract(self):
    #     self.capacity_fulfilled = {key: 0 for key in self.capacity_fulfilled}

    def delete_old_plant_bids(self):
        self.historical_bids.extend(self.accepted_bids)
        self.accepted_bids = []

    def get_upfront_costs(self):
        upfront_cost = self.capacity_mw*(self.pre_dev_cost_per_mw+self.construction_cost_per_mw) + self.infrastructure
        return upfront_cost

    @abstractmethod
    def short_run_marginal_cost(self, model, genco, fuel_price, co2_price):
        pass

    def get_fixed_annual_payments(self):
        pass

    def check_if_operating_in_certain_year(self, current_year, year_difference_from_model_year):
        year_operation_begins = self.get_year_of_operation()
        end_of_life = year_operation_begins + self.operating_period
        year_to_check = year_difference_from_model_year + current_year
        if year_operation_begins < year_to_check <= end_of_life:
            return True
        else:
            return False

    def get_year_of_operation(self):
        year_operation_begins = self.construction_year + self.pre_dev_period + self.construction_period
        return year_operation_begins
    
    def __str__(self):
        ret = "Name: {}. Type: {}. Capacity: {}.".format(self.name, self.plant_type, self.capacity_mw)
        return ret

    def __repr__(self):
        return 'PowerPlant({}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {})'.format(self.name, self.plant_type, self.capacity_mw, self.construction_year, self.average_load_factor, self.pre_dev_period, self.construction_period, self.operating_period, self.pre_dev_spend_years, self.construction_spend_years, self.pre_dev_cost_per_mw, self.construction_cost_per_mw, self.infrastructure, self.fixed_o_and_m_per_mw, self.variable_o_and_m_per_mwh, self.insurance_cost_per_mw, self.connection_cost_per_mw)

